/*
 *   Emu42.c
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "pch.h"
#include "resource.h"
#include "Emu42.h"
#include "io.h"
#include "kml.h"

#define VERSION   "1.02"

#ifdef _DEBUG
LPTSTR szNoTitle = _T("Emu42 ")_T(VERSION)_T(" Debug");
#else
LPTSTR szNoTitle = _T("Emu42 ")_T(VERSION);
#endif
LPTSTR szTitle   = NULL;

static const TCHAR szLicence[] =
	_T("This program is free software; you can redistribute it and/or modify ")
	_T("it under the terms of the GNU General Public License as published by ")
	_T("the Free Software Foundation; either version 2 of the License, or ")
	_T("(at your option) any later version.\r\n")
	_T("\r\n")
	_T("This program is distributed in the hope that it will be useful, ")
	_T("but WITHOUT ANY WARRANTY; without even the implied warranty of ")
	_T("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. ")
	_T("See the GNU General Public License for more details.\r\n")
	_T("\r\n")
	_T("You should have received a copy of the GNU General Public License ")
	_T("along with this program; if not, write to the Free Software Foundation, ")
	_T("Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA");


CRITICAL_SECTION csGDILock;					// critical section for hWindowDC
CRITICAL_SECTION csKeyLock;					// critical section for key scan
CRITICAL_SECTION csT1Lock;					// critical section for timer1 access
CRITICAL_SECTION csT2Lock;					// critical section for timer2 access
CRITICAL_SECTION csSlowLock;				// critical section for speed slow down
INT              nArgc;						// no. of command line arguments
LPCTSTR          *ppArgv;					// command line arguments
LARGE_INTEGER    lFreq;						// high performance counter frequency
LONG             lrcTop;
HANDLE           hHeap;
HANDLE           hThread;
DWORD            lThreadId;
HANDLE           hEventShutdn;				// event handle to stop cpu thread

HINSTANCE        hApp = NULL;
HWND             hWnd = NULL;
HWND             hCmdBar = NULL;
HDC              hWindowDC = NULL;
HPALETTE         hPalette = NULL;
HPALETTE         hOldPalette = NULL;		// old palette of hWindowDC
UINT             nAppZoom = 1;				// application zoom factor
BOOL             bAutoSave = FALSE;
BOOL             bAutoSaveOnExit = TRUE;
BOOL             bAlwaysDisplayLog = TRUE;
BOOL             bLoadObjectWarning = TRUE;


//################
//#
//#    Window Status
//#
//################

VOID SetWindowTitle(LPCTSTR szString)
{
	if (szTitle) HeapFree(hHeap,0,szTitle);

	_ASSERT(hWnd != NULL);
	if (szString)
	{
		LPCTSTR szName;

		// remove path from title
		if ((szName = _tcsrchr(szString,_T('\\'))) != NULL)
		{
			szString = szName + 1;			// cut path before
		}

		szTitle = DuplicateString(szString);
		SetWindowText(hWnd, szTitle);
	}
	else
	{
		szTitle = NULL;
		SetWindowText(hWnd, szNoTitle);
	}
	return;
}



//################
//#
//#    Settings
//#
//################

static BOOL CALLBACK SettingsProc(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	switch (message)
	{
	case WM_INITDIALOG:
		// init speed checkbox
		SendDlgItemMessage(hDlg,IDC_REALSPEED,BM_SETCHECK,bRealSpeed,0);
		SendDlgItemMessage(hDlg,IDC_AUTOSAVE,BM_SETCHECK,bAutoSave,0);
		SendDlgItemMessage(hDlg,IDC_AUTOSAVEONEXIT,BM_SETCHECK,bAutoSaveOnExit,0);
		SendDlgItemMessage(hDlg,IDC_OBJECTLOADWARNING,BM_SETCHECK,bLoadObjectWarning,0);
		SendDlgItemMessage(hDlg,IDC_ALWAYSDISPLOG,BM_SETCHECK,bAlwaysDisplayLog,0);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			// get speed checkbox value
			bRealSpeed = SendDlgItemMessage(hDlg,IDC_REALSPEED,BM_GETCHECK,0,0);
			bAutoSave = SendDlgItemMessage(hDlg,IDC_AUTOSAVE,BM_GETCHECK,0,0);
			bAutoSaveOnExit = SendDlgItemMessage(hDlg,IDC_AUTOSAVEONEXIT,BM_GETCHECK,0,0);
			bLoadObjectWarning = SendDlgItemMessage(hDlg,IDC_OBJECTLOADWARNING,BM_GETCHECK,0,0);
			bAlwaysDisplayLog = SendDlgItemMessage(hDlg,IDC_ALWAYSDISPLOG,BM_GETCHECK,0,0);
			SetSpeed(bRealSpeed);			// set speed
		case IDCANCEL:
			EndDialog(hDlg, LOWORD(wParam));
		}
		break;
	}
	return FALSE;
	UNREFERENCED_PARAMETER(lParam);
}


//################
//#
//#    Save Helper
//#
//################

//
// UINT SaveChanges(BOOL bAuto);
// Return code :
// IDYES    File successfuly saved
// IDNO     File not saved
// IDCANCEL Cancel command
//
static UINT SaveChanges(BOOL bAuto)
{
	UINT uReply;

	if (pbyRom == NULL) return IDNO;

	if (bAuto)
		uReply = IDYES;
	else
		uReply = YesNoCancelMessage(_T("Do you want to save changes?"));

	if (uReply != IDYES) return uReply;

	if (szCurrentFilename[0]==0)
	{ // Save As...
		uReply = GetSaveAsFilename();
		if (uReply != IDOK) return uReply;
		if (!SaveDocumentAs(szBufferFilename)) return IDCANCEL;
		WriteLastDocument(szCurrentFilename);
		return IDYES;
	}

	SaveDocument();
	return IDYES;
}



//################
//#
//#    Message Handlers
//#
//################

//
// WM_CREATE
//
static LRESULT OnCreate(HWND hWindow)
{
	SHMENUBARINFO mbi;

	InitializeCriticalSection(&csGDILock);
	InitializeCriticalSection(&csKeyLock);
	InitializeCriticalSection(&csT1Lock);
	InitializeCriticalSection(&csT2Lock);
	InitializeCriticalSection(&csSlowLock);

	hWnd = hWindow;
	hWindowDC = GetDC(hWnd);

	// create command bar
	ZeroMemory(&mbi,sizeof(SHMENUBARINFO));
	mbi.cbSize     = sizeof (SHMENUBARINFO);
	mbi.hwndParent = hWnd;
	mbi.nToolBarId = IDR_MENU;
	mbi.hInstRes   = hApp;
	SHCreateMenuBar(&mbi);
	hCmdBar = mbi.hwndMB;
	return 0;
}

//
// WM_DESTROY
//
static LRESULT OnDestroy(HWND hWindow)
{
	SwitchToState(SM_RETURN);				// exit emulation thread
	DestroyWindow(hCmdBar);
	SetWindowTitle(NULL);					// free memory of title
	ReleaseDC(hWnd, hWindowDC);
	hCmdBar = NULL;
	hWindowDC = NULL;						// hWindowDC isn't valid any more
	hWnd = NULL;

	DeleteCriticalSection(&csGDILock);
	DeleteCriticalSection(&csKeyLock);
	DeleteCriticalSection(&csT1Lock);
	DeleteCriticalSection(&csT2Lock);
	DeleteCriticalSection(&csSlowLock);

	PostQuitMessage(0);						// exit message loop
	return 0;
	UNREFERENCED_PARAMETER(hWindow);
}

//
// WM_PAINT
//
static LRESULT OnPaint(HWND hWindow)
{
	PAINTSTRUCT Paint;
	HDC hPaintDC;
	DWORD dwScreenFlags;

	if (bFullScreen)
	{
		// workaround for Windows Mobile 2003 (SE) devices
		SHFullScreen(hWindow, SHFS_HIDETASKBAR | SHFS_HIDESIPBUTTON | SHFS_HIDESTARTICON);

		dwScreenFlags = (SHFS_HIDETASKBAR | SHFS_SHOWSIPBUTTON | SHFS_HIDESTARTICON);
	}
	else
	{
		dwScreenFlags = (SHFS_SHOWTASKBAR | SHFS_SHOWSIPBUTTON | SHFS_SHOWSTARTICON);
	}
	SHFullScreen(hWindow, dwScreenFlags);	// change client size

	hPaintDC = BeginPaint(hWindow, &Paint);
	if (hMainDC != NULL)
	{
		INT nxSize,nySize;
		RECT rcMainPaint;

		// redraw position rounded to Zoom
		Paint.rcPaint.left   = Paint.rcPaint.left         / nAppZoom;
		Paint.rcPaint.top    = Paint.rcPaint.top          / nAppZoom;
		Paint.rcPaint.right  = (Paint.rcPaint.right  + 1) / nAppZoom;
		Paint.rcPaint.bottom = (Paint.rcPaint.bottom + 1) / nAppZoom;

		rcMainPaint = Paint.rcPaint;		// coordinates in source bitmap
		rcMainPaint.left   += nBackgroundX;
		rcMainPaint.top    += nBackgroundY;
		rcMainPaint.right  += nBackgroundX;
		rcMainPaint.bottom += nBackgroundY;

		GetSizeLcdBitmap(&nxSize,&nySize);	// get LCD size without AppZoom

		EnterCriticalSection(&csGDILock);	// solving NT GDI problems
		{
			// redraw background bitmap
			StretchBlt(hPaintDC, Paint.rcPaint.left * nAppZoom, Paint.rcPaint.top * nAppZoom,
				       (Paint.rcPaint.right-Paint.rcPaint.left) * nAppZoom,(Paint.rcPaint.bottom-Paint.rcPaint.top) * nAppZoom,
				       hMainDC, rcMainPaint.left, rcMainPaint.top,
				       Paint.rcPaint.right-Paint.rcPaint.left, Paint.rcPaint.bottom-Paint.rcPaint.top,
					   SRCCOPY);
			// redraw display area
			BitBlt(hWindowDC,
				   (nLcdX - nBackgroundX) * nAppZoom, (nLcdY - nBackgroundY) * nAppZoom,
				   nxSize * nAppZoom, nySize * nAppZoom,
				   hLcdDC,
				   0, 0,
				   SRCCOPY);
		}
		LeaveCriticalSection(&csGDILock);
		UpdateAnnunciators();
		RefreshButtons(&rcMainPaint);
	}
	EndPaint(hWindow, &Paint);
	return 0;
}

//
// ID_FILE_NEW
//
static LRESULT OnFileNew(VOID)
{
	SaveBackup();
	if (pbyRom)
	{
		SwitchToState(SM_INVALID);
		if (IDCANCEL == SaveChanges(bAutoSave))
			goto cancel;
	}
	if (NewDocument()) SetWindowTitle(_T("Untitled"));
cancel:
	if (pbyRom) SwitchToState(SM_RUN);
	return 0;
}

//
// ID_FILE_OPEN
//
static LRESULT OnFileOpen(VOID)
{
	if (pbyRom)
	{
		SwitchToState(SM_INVALID);
		if (IDCANCEL == SaveChanges(bAutoSave))
			goto cancel;
	}
	if (GetOpenFilename())
	{
		OpenDocument(szBufferFilename);
	}
cancel:
	if (pbyRom) SwitchToState(SM_RUN);
	return 0;
}

//
// ID_FILE_SAVE
//
static LRESULT OnFileSave(VOID)
{
	if (pbyRom == NULL) return 0;
	SwitchToState(SM_INVALID);
	SaveChanges(TRUE);
	SwitchToState(SM_RUN);
	return 0;
}

//
// ID_FILE_SAVEAS
//
static LRESULT OnFileSaveAs(VOID)
{
	UINT uReply;

	if (pbyRom == NULL) return 0;
	SwitchToState(SM_INVALID);

	uReply = GetSaveAsFilename();
	if (uReply != IDOK)
	{
		SwitchToState(SM_RUN);
		return 0;
	}
	if (!SaveDocumentAs(szBufferFilename))
	{
		SwitchToState(SM_RUN);
		return 0;
	}
	WriteLastDocument(szCurrentFilename);
	SwitchToState(SM_RUN);
	return 0;
}

//
// ID_FILE_CLOSE
//
static LRESULT OnFileClose(VOID)
{
	if (pbyRom == NULL) return 0;
	SwitchToState(SM_INVALID);
	if (SaveChanges(bAutoSave)!=IDCANCEL)
	{
		KillKML();
		ResetDocument();
		SetWindowTitle(NULL);
	}
	else
	{
		SwitchToState(SM_RUN);
	}
	return 0;
}

//
// ID_FILE_EXIT
//
// WM_SYS_CLOSE
//
static LRESULT OnFileExit(VOID)
{
	SwitchToState(SM_INVALID);				// hold emulation thread
	if (SaveChanges(bAutoSaveOnExit) == IDCANCEL)
	{
		SwitchToState(SM_RUN);				// on cancel restart emulation thread
		return 0;
	}
	DestroyWindow(hWnd);
	return 0;
}

//
// ID_STACK_COPY
//
static LRESULT OnStackCopy(VOID)			// copy data from stack
{
	HANDLE hClipObj;
	LPTSTR lpstrClipdata;
	LPBYTE lpbyData;
	DWORD  dwAddress,dwSize;

	_ASSERT(nState == SM_RUN);				// emulator must be in RUN state
	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
	{
		InfoMessage(_T("The emulator is busy."));
		return 0;
	}

	_ASSERT(nState == SM_SLEEP);

	if ((dwAddress = RPL_Pick(1)) == 0)		// pick address of level1 object
	{
		MessageBeep(MB_OK);					// error beep
		goto error;
	}

	if (Read5(dwAddress) != 0x02A4E)		// string object
	{
		MessageBeep(MB_OK);					// error beep
		goto error;
	}

	dwAddress += 5;							// address of string length
	dwSize = (Read5(dwAddress) - 5) / 2;	// length of string

	// memory allocation for clipboard data
	if ((hClipObj = GlobalAlloc(GMEM_MOVEABLE,(dwSize + 1)*sizeof(TCHAR))) == NULL)
		goto error;

	// create temporary buffer for byte data
	if ((lpbyData = HeapAlloc(hHeap,0,dwSize+1)) != NULL)
	{
		DWORD i;

		// copy data into byte buffer
		for (i = 0, dwAddress += 5; i < dwSize; dwAddress += 2)
			lpbyData[i++] = Read2(dwAddress);
		lpbyData[i] = 0;					// set end of string

		if ((lpstrClipdata = GlobalLock(hClipObj)))	// lock memory
		{
			// copy data into clipboard buffer
			MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, lpbyData, dwSize + 1,
								lpstrClipdata, dwSize + 1);

			GlobalUnlock(hClipObj);				// unlock memory

			if (OpenClipboard(hWnd))
			{
				if (EmptyClipboard())
					SetClipboardData(CF_UNICODETEXT,hClipObj);
				else
					GlobalFree(hClipObj);
				CloseClipboard();
			}
		}
		else									// lock memory failed
		{
			GlobalFree(hClipObj);
		}

		HeapFree(hHeap,0,lpbyData);
	}

error:
	SwitchToState(SM_RUN);
	return 0;
}

//
// ID_STACK_PASTE
//
static LRESULT OnStackPaste(VOID)			// paste data to stack
{
	HANDLE  hClipObj;
	LPCTSTR lpstrClipdata;
	LPBYTE  lpbyData;

	BOOL   bSuccess = FALSE;

	if (!(Chipset.IORam[DSPCTL]&DON))		// calculator off, turn on
	{
		KeyboardEvent(TRUE,0,0x8000);
		Sleep(200);
		KeyboardEvent(FALSE,0,0x8000);
		Sleep(200);
		// wait for sleep mode
		while(Chipset.Shutdn == FALSE) Sleep(0);
	}

	_ASSERT(nState == SM_RUN);				// emulator must be in RUN state
	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
	{
		InfoMessage(_T("The emulator is busy."));
		goto cancel;
	}

	_ASSERT(nState == SM_SLEEP);

	if (OpenClipboard(hWnd))
	{
		if (IsClipboardFormatAvailable(CF_UNICODETEXT))
		{
			if ((hClipObj = GetClipboardData(CF_UNICODETEXT)))
			{
				if ((lpstrClipdata = GlobalLock(hClipObj)))
				{
					DWORD dwSize = lstrlen(lpstrClipdata);
					if ((lpbyData = HeapAlloc(hHeap,0,dwSize * 2)))
					{
						// copy data UNICODE -> ASCII
						WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK,
											lpstrClipdata, dwSize,
											lpbyData+dwSize, dwSize, NULL, NULL);
						bSuccess = (WriteStack(lpbyData,dwSize) == S_ERR_NO);
						HeapFree(hHeap,0,lpbyData);
					}
					GlobalUnlock(hClipObj);
				}
			}
		}
		else								// unknown clipboard format
		{
			MessageBeep(MB_OK);				// error beep
		}
		CloseClipboard();
	}

	SwitchToState(SM_RUN);					// run state
	while (nState!=nNextState) Sleep(0);
	_ASSERT(nState == SM_RUN);

	if (bSuccess == FALSE)					// data not copied
		goto cancel;

	KeyboardEvent(TRUE,0,0x8000);
	Sleep(200);
	KeyboardEvent(FALSE,0,0x8000);

	// wait for sleep mode
	while(Chipset.Shutdn == FALSE) Sleep(0);

cancel:
	return 0;
}

//
// ID_VIEW_RESET
//
static LRESULT OnViewReset(VOID)
{
	if (nState != SM_RUN) return 0;
	if (YesNoMessage(_T("Are you sure you want to press the Reset Button?"))==IDYES)
	{
		SwitchToState(SM_SLEEP);
		CpuReset();							// register setting after Cpu Reset
		SwitchToState(SM_RUN);
	}
	return 0;
}

//
// ID_VIEW_SETTINGS
//
static LRESULT OnViewSettings(VOID)
{
	UINT nOldState = nState;
	
	if (nState == SM_RUN)					// currently running
		SwitchToState(SM_SLEEP);			// suspend emulation

	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_SETTINGS), hWnd, (DLGPROC)SettingsProc) == -1)
		AbortMessage(_T("Settings Dialog Creation Error!"));

	WriteSettings();						// update INI file
	SwitchToState(nOldState);				// switch back to old state
	return 0;
}

//
// ID_VIEW_SCRIPT
//
static LRESULT OnViewScript(VOID)
{
	BYTE cType = cCurrentRomType;
	if (nState != SM_RUN)
	{
		InfoMessage(_T("You cannot change the KML script when Emu42 is not running.\n")
					_T("Use the File,New menu item to create a new calculator."));
		return 0;
	}
	SwitchToState(SM_INVALID);
retry:
	if (!DisplayChooseKml(cType)) goto ok;
	if (!InitKML(szCurrentKml,FALSE))
	{
		if (AbortMessage(_T("This KML Script has errors.\nPress Ok to try to find a correct script."))==IDCANCEL)
		{
			if (szCurrentFilename[0])
				OnFileSave();
			else
				OnFileSaveAs();
			goto ok;
		}
		goto retry;
	}
ok:
	if (pbyRom) SwitchToState(SM_RUN);
	return 0;
}

//
// ID_BACKUP_SAVE
//
static LRESULT OnBackupSave(VOID)
{
	UINT nOldState;
	if (pbyRom == NULL) return 0;
	nOldState = SwitchToState(SM_INVALID);
	SaveBackup();
	SwitchToState(nOldState);
	return 0;
}

//
// ID_BACKUP_RESTORE
//
static LRESULT OnBackupRestore(VOID)
{
	SwitchToState(SM_INVALID);
	RestoreBackup();
	if (pbyRom) SwitchToState(SM_RUN);
	return 0;
}

//
// ID_BACKUP_DELETE
//
static LRESULT OnBackupDelete(VOID)
{
	ResetBackup();
	return 0;
}

//
// ID_OBJECT_LOAD
//
static LRESULT OnObjectLoad(VOID)
{
	if (!(Chipset.IORam[DSPCTL]&DON))		// calculator off, turn on
	{
		KeyboardEvent(TRUE,0,0x8000);
		Sleep(200);
		KeyboardEvent(FALSE,0,0x8000);
		Sleep(200);
		// wait for sleep mode
		while(Chipset.Shutdn == FALSE) Sleep(0);
	}

	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to load an object."));
		goto cancel;
	}

	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
	{
		InfoMessage(_T("The emulator is busy."));
		goto cancel;
	}

	_ASSERT(nState == SM_SLEEP);

	if (bLoadObjectWarning)
	{
		UINT uReply = YesNoCancelMessage(
			_T("Warning: Trying to load an object while the emulator is busy ")
			_T("will certainly result in a memory lost. Before loading an object ")
			_T("you should be sure that the calculator is not doing anything. ")
			_T("Do you want to see this warning next time you try to load an object?\n"));
		switch (uReply)
		{
		case IDYES:
			break;
		case IDNO:
			bLoadObjectWarning = FALSE;
			break;
		case IDCANCEL:
			SwitchToState(SM_RUN);
			goto cancel;
		}
	}

	if (!GetLoadObjectFilename())
	{
		SwitchToState(SM_RUN);
		goto cancel;
	}

	if (cCurrentRomType == 'D')				// HP42S
	{
		GetUserCode(szBufferFilename);
		SwitchToState(SM_RUN);
	}
	else									// HP28S
	{
		if (!LoadObject(szBufferFilename))
		{
			SwitchToState(SM_RUN);
			goto cancel;
		}

		SwitchToState(SM_RUN);				// run state
		while (nState!=nNextState) Sleep(0);
		_ASSERT(nState == SM_RUN);
		KeyboardEvent(TRUE,0,0x8000);
		Sleep(200);
		KeyboardEvent(FALSE,0,0x8000);
		while(Chipset.Shutdn == FALSE) Sleep(0);
	}

cancel:
	return 0;
}

//
// ID_OBJECT_SAVE
//
static LRESULT OnObjectSave(VOID)
{
	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to save an object."));
		return 0;
	}

	if (WaitForSleepState())				// wait for cpu SHUTDN then sleep state
	{
		InfoMessage(_T("The emulator is busy."));
		return 0;
	}
	_ASSERT(nState == SM_SLEEP);

	if (cCurrentRomType == 'D')				// HP42S
	{
		OnSelectProgram();
	}
	else									// HP28S
	{
		_ASSERT(cCurrentRomType == 'O');
		if (!GetSaveObjectFilename())
		{
			SwitchToState(SM_RUN);
			return 0;
		}

		SaveObject(szBufferFilename);
	}

	SwitchToState(SM_RUN);
	return 0;
}

//
// ID_ABOUT
//
static BOOL CALLBACK About(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	switch (message)
	{
	case WM_INITDIALOG:
		SetDlgItemText(hDlg,IDC_VERSION,szNoTitle);
		SetDlgItemText(hDlg,IDC_LICENSE,szLicence);
		FULLSCREEN(hDlg);
		return TRUE;
	case WM_COMMAND:
		wParam = LOWORD(wParam);
		if ((wParam==IDOK)||(wParam==IDCANCEL))
		{
			EndDialog(hDlg, wParam);
			return TRUE;
		}
		break;
	}
	return FALSE;
	UNREFERENCED_PARAMETER(lParam);
}

static LRESULT OnAbout(VOID)
{
	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_ABOUT), hWnd, (DLGPROC)About) == -1)
		AbortMessage(_T("About Dialog Box Creation Error!"));
	return 0;
}

static LRESULT OnPopupMenu(HMENU hMenu,UINT uPos)
{
	// enable stack loading on HP28S and User Code loading on HP42S
	BOOL bObjectEnable = (cCurrentRomType == 'D') || (cCurrentRomType == 'O');
	BOOL bStackEnable  = (cCurrentRomType == 'O');
	BOOL bRun          = nState == SM_RUN || nState == SM_SLEEP;

	UINT uObjectEnable = (bRun && bObjectEnable) ? MF_ENABLED : MF_GRAYED;
	UINT uStackEnable  = (bRun && bStackEnable)  ? MF_ENABLED : MF_GRAYED;
	UINT uRun          = bRun                    ? MF_ENABLED : MF_GRAYED;
	UINT uBackup       = bBackup                 ? MF_ENABLED : MF_GRAYED;

	switch (uPos)
	{
	case ID_FILE:
		EnableMenuItem(hMenu,ID_FILE_NEW,MF_ENABLED);
		EnableMenuItem(hMenu,ID_FILE_OPEN,MF_ENABLED);
		EnableMenuItem(hMenu,ID_FILE_SAVE,(bRun && szCurrentFilename[0]) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(hMenu,ID_FILE_SAVEAS,uRun);
		EnableMenuItem(hMenu,ID_FILE_CLOSE,uRun);
		break;
	case ID_EDIT:
		EnableMenuItem(hMenu,ID_OBJECT_LOAD,uObjectEnable);
		EnableMenuItem(hMenu,ID_OBJECT_SAVE,uObjectEnable);
		EnableMenuItem(hMenu,ID_STACK_COPY,uStackEnable);
		EnableMenuItem(hMenu,ID_STACK_PASTE,uStackEnable);
		EnableMenuItem(hMenu,ID_VIEW_RESET,uRun);
		EnableMenuItem(hMenu,ID_BACKUP_SAVE,uRun);
		EnableMenuItem(hMenu,ID_BACKUP_RESTORE,uBackup);
		EnableMenuItem(hMenu,ID_BACKUP_DELETE,uBackup);
		break;
	case ID_VIEW:
		EnableMenuItem(hMenu,ID_VIEW_SCRIPT,uRun);
		break;
	}
	return 0;
}

static LRESULT OnLButtonDown(UINT nFlags, WORD x, WORD y)
{
	if (nState == SM_RUN) MouseButtonDownAt(nFlags, x,y);
	return 0;
}

static LRESULT OnLButtonUp(UINT nFlags, WORD x, WORD y)
{
	if (nState == SM_RUN) MouseButtonUpAt(nFlags, x,y);
	return 0;
}

static LRESULT OnMouseMove(UINT nFlags, WORD x, WORD y)
{
	if (nState == SM_RUN) MouseMovesTo(nFlags, x,y);
	return 0;
}

static LRESULT OnKeyDown(int nVirtKey, DWORD lKeyData)
{
	// call RunKey() only once (suppress autorepeat feature)
	if (nState == SM_RUN && (lKeyData & 0x40000000) == 0)
		RunKey((BYTE)nVirtKey, TRUE);
	return 0;
}

static LRESULT OnKeyUp(int nVirtKey, DWORD lKeyData)
{
	if (nState == SM_RUN) RunKey((BYTE)nVirtKey, FALSE);
	return 0;
	UNREFERENCED_PARAMETER(lKeyData);
}

LRESULT CALLBACK MainWndProc(HWND hWindow, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_CREATE:          return OnCreate(hWindow);
	case WM_DESTROY:         return OnDestroy(hWindow);
	case WM_PAINT:           return OnPaint(hWindow);
	case WM_ACTIVATE:
		if (LOWORD(wParam)==WA_INACTIVE) break;
	case WM_QUERYNEWPALETTE:
		if (hPalette)
		{
			SelectPalette(hWindowDC, hPalette, FALSE);
			if (RealizePalette(hWindowDC))
			{
				InvalidateRect(hWindow,NULL,TRUE);
				return TRUE;
			}
		}
		return FALSE;
	case WM_PALETTECHANGED:
		if ((HWND)wParam == hWindow) break;
		if (hPalette)
		{
			SelectPalette(hWindowDC, hPalette, FALSE);
			if (RealizePalette(hWindowDC))
			{
				// UpdateColors(hWindowDC);
				InvalidateRect (hWnd, (LPRECT) (NULL), 1);
			}
		}
		return FALSE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case ID_FILE_NEW:      return OnFileNew();
		case ID_FILE_OPEN:     return OnFileOpen();
		case ID_FILE_SAVE:     return OnFileSave();
		case ID_FILE_SAVEAS:   return OnFileSaveAs();
		case ID_FILE_CLOSE:    return OnFileClose();
		case ID_FILE_EXIT:     return OnFileExit();
		case ID_STACK_COPY:    return OnStackCopy();
		case ID_STACK_PASTE:   return OnStackPaste();
		case ID_VIEW_RESET:    return OnViewReset();
		case ID_VIEW_SETTINGS: return OnViewSettings();
		case ID_VIEW_SCRIPT:   return OnViewScript();
		case ID_BACKUP_SAVE:   return OnBackupSave();
		case ID_BACKUP_RESTORE:return OnBackupRestore();
		case ID_BACKUP_DELETE: return OnBackupDelete();
		case ID_OBJECT_LOAD:   return OnObjectLoad();
		case ID_OBJECT_SAVE:   return OnObjectSave();
		case ID_ABOUT:         return OnAbout();
		}
		break;
	case WM_SYSCOMMAND:
		switch (wParam & 0xFFF0)
		{
		case SC_CLOSE: return OnFileExit();
		}
		break;

	case WM_INITMENUPOPUP: return OnPopupMenu((HMENU)wParam, LOWORD(lParam));
	case WM_RBUTTONDOWN:
	case WM_LBUTTONDOWN:   return OnLButtonDown(wParam, LOWORD(lParam), HIWORD(lParam));
	case WM_LBUTTONUP:     return OnLButtonUp(wParam, LOWORD(lParam), HIWORD(lParam));
	case WM_MOUSEMOVE:     return OnMouseMove(wParam, LOWORD(lParam), HIWORD(lParam));
	case WM_KEYUP:         return OnKeyUp((int)wParam, lParam);
	case WM_KEYDOWN:       return OnKeyDown((int)wParam, lParam);
	}
	return DefWindowProc(hWindow, uMsg, wParam, lParam);
}

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPTSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
	WNDCLASS wc;
	RECT rc;

	hApp = hInst;
	nArgc = __argc;						// no. of command line arguments
	ppArgv = (LPCTSTR*) __argv;			// command line arguments

	// running an instance of me?
	if ((hWnd = FindWindow(_T("CEmu42"),NULL)) != NULL)
	{
		SetForegroundWindow(hWnd);		// switch to it
		return FALSE;					// quit
	}

	SHInitExtraControls();

	if(!QueryPerformanceFrequency(&lFreq))	// init high resolution counter
	{
		AbortMessage(
			_T("No high resolution timer available.\n")
			_T("This application will now terminate."));
		return FALSE;
	}

	hHeap = GetProcessHeap();
	if (hHeap == NULL)
	{
		AbortMessage(_T("Heap creation failed."));
		return FALSE;
	}

	memset(&wc,0,sizeof(WNDCLASS));
	wc.lpfnWndProc = (WNDPROC)MainWndProc;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_EMU42));
	wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszClassName = _T("CEmu42");

	if (!RegisterClass(&wc))
	{
		AbortMessage(
			_T("CEmu42 class registration failed.\n")
			_T("This application will now terminate."));
		return FALSE;
	}

	// Create window
	hWnd = CreateWindow(_T("CEmu42"), _T("Emu42"),
		WS_VISIBLE,
		CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT,CW_USEDEFAULT,
		NULL,NULL,hApp,NULL
		);

	if (hWnd == NULL)
	{
		AbortMessage(_T("Window creation failed."));
		return FALSE;
	}

	GetWindowRect(hWnd, &rc);				// get current size of top menu
	lrcTop = rc.top;

	// initialization
	ReadSettings();

	// create auto event handle
	hEventShutdn = CreateEvent(NULL,FALSE,FALSE,NULL);
	if (hEventShutdn == NULL)
	{
		AbortMessage(_T("Event creation failed."));
		DestroyWindow(hWnd);
		return FALSE;
	}

	nState     = SM_RUN;					// init state must be <> nNextState
	nNextState = SM_INVALID;				// go into invalid state
	hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)&WorkerThread, NULL, 0, &lThreadId);
	if (hThread == NULL)
	{
		CloseHandle(hEventShutdn);			// close event handle
		AbortMessage(_T("Thread creation failed."));
		DestroyWindow(hWnd);
		return FALSE;
	}
	while (nState!=nNextState) Sleep(0);	// wait for thread initialized

	_ASSERT(hWnd != NULL);
	_ASSERT(hWindowDC != NULL);

	if (nArgc >= 2)							// use decoded parameter line
	{
		TCHAR szTemp[MAX_PATH+8] = _T("Loading ");

		// remove path from title
		LPCTSTR szName = _tcsrchr(ppArgv[1],_T('\\'));
		lstrcat(szTemp, (szName != NULL) ? (szName + 1) : ppArgv[1]);

		SetWindowTitle(szTemp);
		if (OpenDocument(ppArgv[1]))
			goto start;
	}

	ReadLastDocument(szBufferFilename, ARRAYSIZEOF(szBufferFilename));
	if (szBufferFilename[0])
	{
		TCHAR szTemp[MAX_PATH+8] = _T("Loading ");

		// remove path from title
		LPCTSTR szName = _tcsrchr(szBufferFilename,_T('\\'));
		lstrcat(szTemp, (szName != NULL) ? (szName + 1) : szBufferFilename);

		SetWindowTitle(szTemp);
		if (OpenDocument(szBufferFilename))
			goto start;
	}

	SetWindowTitle(_T("New Document"));
	if (NewDocument())
	{
		SetWindowTitle(_T("Untitled"));
		goto start;
	}

	ResetDocument();

start:
	ShowWindow(hWnd, nCmdShow);

	if (pbyRom) SwitchToState(SM_RUN);

	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	WriteSettings();						// save emulation settings

	CloseHandle(hThread);					// close thread handle
	CloseHandle(hEventShutdn);				// close event handle
	_ASSERT(nState == SM_RETURN);			// emulation thread down?
	ResetDocument();
	ResetBackup();
	_ASSERT(pKml == NULL);					// KML script not closed
	_ASSERT(szTitle == NULL);				// freed allocated memory
	_ASSERT(hPalette == NULL);				// freed resource memory

	return msg.wParam;
	UNREFERENCED_PARAMETER(lpCmdLine);
	UNREFERENCED_PARAMETER(hPrevInst);
}
