/*
 *   Emu42.h
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "types.h"

#define	ARRAYSIZEOF(a)	(sizeof(a) / sizeof(a[0]))

#define	_KB(a)			((a)*2*1024)

#define HARDWARE		"Lewis"				// emulator hardware

#define BINARYHEADER28  "HPHP28-W"

#define T2CYCLES		(dwLewisCyclesRate)	// CPU cycles in 16384 Hz time frame

#define SM_RUN			0					// states of cpu emulation thread
#define SM_INVALID		1
#define SM_RETURN		2
#define SM_SLEEP		3

#define S_ERR_NO        0					// stack errorcodes
#define S_ERR_BINARY    1
#define S_ERR_ASCII     2

// values for selection of Lewis chip
enum CHIP { MASTER = 0, SLAVE };

// values for mapping area
enum MMUMAP { M_NONE,
			  M_REG,  M_DISP,  M_NCE2,  M_NCE3,  M_ROM,
			  M_REGS, M_DISPS, M_NCE2S, M_NCE3S, M_ROMS
			};

// Emu42.c
extern HPALETTE			hPalette;
extern HPALETTE			hOldPalette;
extern HANDLE			hEventShutdn;
extern LPTSTR			szTitle;
extern CRITICAL_SECTION	csGDILock;
extern CRITICAL_SECTION	csKeyLock;
extern CRITICAL_SECTION	csT1Lock;
extern CRITICAL_SECTION	csT2Lock;
extern CRITICAL_SECTION csSlowLock;
extern INT				nArgc;
extern LPCTSTR			*ppArgv;
extern LARGE_INTEGER	lFreq;
extern LONG				lrcTop;
extern HANDLE			hHeap;
extern HINSTANCE		hApp;
extern HWND				hWnd;
extern HDC				hWindowDC;
extern UINT             nAppZoom;
extern BOOL				bAutoSave;
extern BOOL				bAutoSaveOnExit;
extern BOOL				bAlwaysDisplayLog;
extern BOOL				bLoadObjectWarning;
extern HANDLE			hThread;
extern DWORD			lThreadId;
extern VOID				SetWindowTitle(LPCTSTR szString);

// Settings.c
extern VOID ReadSettings(VOID);
extern VOID WriteSettings(VOID);
extern VOID ReadLastDocument(LPTSTR szFileName, DWORD nSize);
extern VOID WriteLastDocument(LPCTSTR szFilename);

// Display.c
extern BOOL   bFullScreen;
extern BOOL   bVertical;
extern UINT   nBackgroundX;
extern UINT   nBackgroundY;
extern UINT   nBackgroundW;
extern UINT   nBackgroundH;
extern UINT   nLcdX;
extern UINT   nLcdY;
extern UINT   nLcdZoom;
extern LPBYTE pbyLcd;
extern HDC    hLcdDC;
extern HDC    hMainDC;
extern VOID   UpdateContrast(BYTE byContrast);
extern VOID   SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue);
extern VOID   GetSizeLcdBitmap(INT *pnX, INT *pnY);
extern VOID   CreateLcdBitmap(VOID);
extern VOID   DestroyLcdBitmap(VOID);
extern BOOL   CreateMainBitmap(LPCTSTR szFilename);
extern VOID   DestroyMainBitmap(VOID);
extern VOID   UpdateAnnunciators(VOID);
extern VOID   StartDisplay(VOID);
extern VOID   StopDisplay(VOID);
extern VOID   ResizeWindow(VOID);

// Engine.c
extern CHIPSET_M Chipset;
extern CHIPSET_S ChipsetS;
extern BOOL      bInterrupt;
extern UINT      nState;
extern UINT      nNextState;
extern BOOL      bRealSpeed;
extern BOOL      bKeySlow;
extern DWORD     dwLewisCycles;
extern DWORD     dwLewisCyclesRate;
extern VOID      AdjKeySpeed(VOID);
extern VOID      SetSpeed(BOOL bAdjust);
extern BOOL      WaitForSleepState(VOID);
extern UINT      SwitchToState(UINT nNewState);
extern UINT      WorkerThread(LPVOID pParam);

// Fetch.c
extern VOID    EvalOpcode(LPBYTE I);

// Files.c
extern TCHAR   szEmuDirectory[MAX_PATH];
extern TCHAR   szDocDirectory[MAX_PATH];
extern TCHAR   szCurrentKml[MAX_PATH];
extern TCHAR   szBackupKml[MAX_PATH];
extern TCHAR   szCurrentFilename[MAX_PATH];
extern TCHAR   szBackupFilename[MAX_PATH];
extern TCHAR   szBufferFilename[MAX_PATH];
extern LPBYTE  pbyRom;
extern DWORD   dwRomSize;
extern BYTE    cCurrentRomType;
extern UINT    nCurrentClass;
extern BOOL    bBackup;
extern WORD    WriteStack(LPBYTE,DWORD);
extern LPCTSTR FullFilename(LPCTSTR lpszFilename);
extern BOOL    PatchRom(LPCTSTR szFilename);
extern BOOL    MapRom(LPCTSTR szFilename);
extern VOID    UnmapRom(VOID);
extern VOID    ResetDocument(VOID);
extern BOOL    NewDocument(VOID);
extern BOOL    OpenDocument(LPCTSTR szFilename);
extern BOOL    SaveDocument(VOID);
extern BOOL    SaveDocumentAs(LPCTSTR szFilename);
extern BOOL    SaveBackup(VOID);
extern BOOL    RestoreBackup(VOID);
extern BOOL    ResetBackup(VOID);
extern BOOL    GetOpenFilename(VOID);
extern BOOL    GetSaveAsFilename(VOID);
extern BOOL    GetLoadObjectFilename(VOID);
extern BOOL    GetSaveObjectFilename(VOID);
extern BOOL    LoadObject(LPCTSTR szFilename);
extern BOOL    SaveObject(LPCTSTR szFilename);
extern HBITMAP LoadBitmapFile(LPCTSTR szFilename);

// Timer.c
extern VOID  SetHPTime(VOID);
extern VOID  StartTimers(enum CHIP eChip);
extern VOID  StopTimers(enum CHIP eChip);
extern DWORD ReadT2(enum CHIP eChip);
extern VOID  SetT2(enum CHIP eChip, DWORD dwValue);
extern BYTE  ReadT1(enum CHIP eChip);
extern VOID  SetT1(enum CHIP eChip, BYTE byValue);

// MOps.c
extern LPBYTE      RMap[1024];
extern LPBYTE      WMap[1024];
extern WORD        UpCRC(WORD wCRC, BYTE nib);
extern VOID        Map(WORD a, WORD b);
extern VOID        Config(VOID);
extern VOID        Uncnfg(VOID);
extern enum MMUMAP MapData(DWORD d);
extern VOID        Reset(VOID);
extern VOID        C_Eq_Id(VOID);
extern VOID        CpuReset(VOID);
extern VOID        Npeek(BYTE *a, DWORD d, UINT s);
extern VOID        Nread(BYTE *a, DWORD d, UINT s);
extern VOID        Nwrite(BYTE *a, DWORD d, UINT s);
extern BYTE        Read2(DWORD d);
extern DWORD       Read5(DWORD d);
extern VOID        Write5(DWORD d, DWORD n);
extern VOID        Write2(DWORD d, BYTE n);
extern VOID        ReadIO(BYTE *a, DWORD b, DWORD s, BOOL bUpdate);
extern VOID        WriteIO(BYTE *a, DWORD b, DWORD s);
extern VOID        ReadSlaveIO(BYTE *a, DWORD d, DWORD s, BOOL bUpdate);
extern VOID        WriteSlaveIO(BYTE *a, DWORD d, DWORD s);

// Keyboard.c
extern VOID ScanKeyboard(BOOL bActive, BOOL bReset);
extern VOID KeyboardEvent(BOOL bPress, UINT out, UINT in);

// Usrprg42.c
extern BOOL GetUserCode(LPCTSTR szUCFile);
extern BOOL OnSelectProgram(VOID);

// RPL.c
extern DWORD RPL_SkipOb(DWORD d);
extern DWORD RPL_ObjectSize(BYTE *o);
extern DWORD RPL_CreateTemp(DWORD l);
extern DWORD RPL_Pick(UINT l);
extern VOID  RPL_Replace(DWORD n);
extern VOID  RPL_Push(DWORD n);

// Message Boxes
static __inline int InfoMessage(LPCTSTR szMessage)  {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_OK|MB_ICONINFORMATION|MB_SETFOREGROUND);}
static __inline int AbortMessage(LPCTSTR szMessage) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_OK|MB_ICONSTOP|MB_SETFOREGROUND);}
static __inline int YesNoMessage(LPCTSTR szMessage) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_YESNO|MB_ICONEXCLAMATION|MB_SETFOREGROUND);}
static __inline int YesNoCancelMessage(LPCTSTR szMessage) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_YESNOCANCEL|MB_ICONEXCLAMATION|MB_SETFOREGROUND);}

// Missing Win32 API calls
static __inline LPTSTR DuplicateString(LPCTSTR szString)
{
	UINT   uLength = lstrlen(szString) + 1;
	LPTSTR szDup   = HeapAlloc(hHeap,0,uLength*sizeof(szDup[0]));
	lstrcpy(szDup,szString);
	return szDup;
}

// Full screen dialog box
static __inline VOID FULLSCREEN(HWND hDlg)
{
	SHINITDLGINFO shidi;

	_ASSERT(hWnd != NULL);
	SHFullScreen(hWnd, SHFS_SHOWTASKBAR | SHFS_SHOWSIPBUTTON | SHFS_SHOWSTARTICON);

	shidi.dwMask = SHIDIM_FLAGS;
	shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIZEDLG;
	shidi.hDlg = hDlg;
	SHInitDialog(&shidi);
	return;
}

// WinCE replacement
static __inline DWORD GetCurrentDirectory(DWORD nBufferLength,LPTSTR lpBuffer)
{
	LPTSTR lpPos;

	// get path and name of current file
	GetModuleFileName(NULL,lpBuffer,nBufferLength);

	// get delimiter position of filename
	if ((lpPos = _tcsrchr(lpBuffer,_T('\\'))) != NULL)
	{
		*lpPos = 0;							// purge filename
	}
	else
	{
		lpBuffer[0] = _T('\\');				// root path
		lpBuffer[1] = 0;
	}
	return lstrlen(lpBuffer);
}
