/*
 *   rpl.c
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "pch.h"
#include "Emu42.h"
#include "ops.h"
#include "io.h"

//| 28S  | 42S  | Name
//#C005A #500B8 =TEMPTOP
//#C005F #500BD =RSKTOP   (B)
//#C0064 #500C2 =DSKTOP   (D1)
//#C00DC #50117 =AVMEM    (D)
//#C008C #500E0 =INTRPPTR (D0)

// HP28S entries
#define TEMPTOP		0xC005A
#define RSKTOP		0xC005F
#define DSKTOP		0xC0064
#define AVMEM		0xC00DC
#define INTRPPTR	0xC008C

#define DOBINT		0x02911		// System Binary
#define DOREAL		0x02933		// Real
#define DOEREAL		0x02955		// Long Real
#define DOCMP		0x02977		// Complex
#define DOECMP		0x0299D		// Long Complex
#define DOCHAR		0x029BF		// Character
#define DOEXT		0x029E1		// Reserved
#define DOARRY		0x02A0A		// Array
#define DOLNKARRY	0x02A2C		// Linked Array
#define DOCSTR		0x02A4E		// String
#define DOHSTR		0x02A70		// Binary Integer
#define DOLIST		0x02A96		// List
#define DORRP		0x02AB8		// Directory
#define DOSYMB		0x02ADA		// Algebraic
#define DOCOL		0x02C67		// Program
#define DOCODE		0x02C96		// Code
#define DOIDNT		0x02D12		// Global Name
#define DOLAM		0x02D37		// Local Name
#define DOROMP		0x02D5C		// XLIB Name
#define SEMI		0x02F90		// ;

DWORD RPL_SkipOb(DWORD d)
{
	BYTE X[8];
	DWORD n, l;

	Npeek(X,d,5);
	n = Npack(X, 5); // read prolog
	switch (n)
	{
	case DOBINT:  l = 10; break; // System Binary
	case DOREAL:  l = 21; break; // Real
	case DOEREAL: l = 26; break; // Long Real
	case DOCMP:   l = 37; break; // Complex
	case DOECMP:  l = 47; break; // Long Complex
	case DOCHAR:  l =  7; break; // Character
	case DOROMP:  l = 11; break; // XLIB Name
	case DOLIST: // List
	case DOSYMB: // Algebraic
	case DOCOL:  // Program
		n=d+5;
		do
		{
			d=n; n=RPL_SkipOb(d);
		} while (d!=n);
		return n+5;
	case SEMI: return d; // ;
	case DOIDNT: // Global Name
	case DOLAM:  // Local Name
		Npeek(X,d+5,2); n = 7 + Npack(X,2)*2;
		return RPL_SkipOb(d+n);
	case DORRP: // Directory
		d+=8;
		n = Read5(d);
		if (n==0)
		{
			return d+5;
		}
		else
		{
			d+=n;
			Npeek(X,d,2);
			n = Npack(X,2)*2 + 4;
			return RPL_SkipOb(d+n);
		}
	case DOEXT:     // Reserved
	case DOARRY:    // Array
	case DOLNKARRY: // Linked Array
	case DOCSTR:    // String
	case DOHSTR:    // Binary Integer
	case DOCODE:    // Code
		l = 5+Read5(d+5);
		break;
	default: return d+5;
	}
	return d+l;
}

DWORD RPL_ObjectSize(BYTE *o)
{
	DWORD n, l = 0;

	n = Npack(o, 5); // read prolog
	switch (n)
	{
	case DOBINT:  l = 10; break; // System Binary
	case DOREAL:  l = 21; break; // Real
	case DOEREAL: l = 26; break; // Long Real
	case DOCMP:   l = 37; break; // Complex
	case DOECMP:  l = 47; break; // Long Complex
	case DOCHAR:  l =  7; break; // Character
	case DOROMP:  l = 11; break; // XLIB Name
	case DOLIST: // List
	case DOSYMB: // Algebraic
	case DOCOL:  // Program
		n=5;
		do { l+=n; o+=n; n=RPL_ObjectSize(o); } while (n);
		l += 5;
		break;
	case SEMI: l =  0; break; // ;
	case DOIDNT: // Global Name
	case DOLAM:  // Local Name
		n = 7 + Npack(o+5,2)*2;
		l = n + RPL_ObjectSize(o+n);
		break;
	case DORRP: // Directory
		n = Npack(o+8,5);
		if (n==0) // empty dir
		{
			l=13;
		}
		else
		{
			l = 8+n;
			n = Npack(o+l,2)*2 + 4;
			l += n;
			l += RPL_ObjectSize(o+l);
		}
		break;
	case DOEXT:     // Reserved
	case DOARRY:    // Array
	case DOLNKARRY: // Linked Array
	case DOCSTR:    // String
	case DOHSTR:    // Binary Integer
	case DOCODE:    // Code
		l = 5 + Npack(o+5,5);
		break;
	default: l=5;
	}
	return l;
}

DWORD RPL_CreateTemp(DWORD l)
{
	DWORD a, b, c;
	BYTE *p;

	l += 6;									// memory for link field (5) + marker (1) and end
	a = Read5(TEMPTOP);						// tail address of top object
	b = Read5(RSKTOP);						// tail address of rtn stack
	c = Read5(DSKTOP);						// top of data stack
	if ((b+l)>c) return 0;					// check if there's enough memory to move DSKTOP
	Write5(TEMPTOP, a+l);					// adjust new end of top object
	Write5(RSKTOP,  b+l);					// adjust new end of rtn stack
	Write5(AVMEM, (c-b-l)/5);				// calculate free memory (*5 nibbles)
	p = HeapAlloc(hHeap,0,b-a);				// move down rtn stack
	Npeek(p,a,b-a);
	Nwrite(p,a+l,b-a);
	HeapFree(hHeap,0,p);
	Write5(a+l-5,l);						// set object length field
	return (a+1);							// return base address of new object
}

DWORD RPL_Pick(UINT l)
{
	DWORD stkp;

	_ASSERT(l > 0);							// first stack elememt is one
	if (l==0) return 0;
	stkp = Read5(DSKTOP) + (l-1)*5;
	return Read5(stkp);
}

#if 0										// function not needed yet
VOID RPL_Replace(DWORD n)
{
	DWORD stkp;

	stkp = Read5(DSKTOP);
	Write5(stkp,n);
	return;
}
#endif

VOID RPL_Push(DWORD n)
{
	DWORD stkp, avmem;

	avmem = Read5(AVMEM);					// amount of free memory
	if (avmem==0) return;					// no memory free
	avmem--;								// fetch memory
	Write5(AVMEM,avmem);					// save new amount of free memory
	stkp = Read5(DSKTOP);					// get pointer to stack level 1
	stkp-=5;								// fetch new stack entry
	Write5(stkp,n);							// save pointer to new object on stack level 1
	Write5(DSKTOP,stkp);					// save new pointer to stack level 1
	return;
}
